//
// Bmp2Cvi by ssjx (http://ssjx.co.uk)
//
// v2.5 C# Framework
// + Can now handle larger images!
//

using System;
using System.IO;

class Bmp2cvi
{
	public struct Mybmpinfo{
		public int width;
		public int height;
		public short depth;
	};

	static byte popular(byte[] pix){
		int[] count=new int[256];

		foreach(var px in pix){
			count[px]++;
		}

		// find largest
		int big=count[0];  // number of occurences
		int bigc=0;

		for(int c=1;c<count.Length;c++){
			if (count[c]>big){
				big=count[c];  // number of occurences
				bigc=c;        // actual value
			}
		}
		
		return (byte)bigc;
	}
	//

	static int swp(byte a,byte b,byte c,byte d){
		int ta = a;
		int tb = b;
		int tc = c;
		int td = d;
		return ta + (tb << 8)+ (tc << 16)+ (td << 24);
	}
	
	static int byte_swap(byte a,byte b){
		int ta = a;
		int tb = b;
		return ta + (tb << 8);
	}
		
	static void Main(string[] args)
	{	
		string fn="";
		int dither=0;
		
		Console.WriteLine("Bmp2Cvi v2.5 C# (22/05/23) by ssjx (http://ssjx.co.uk)");
		Console.WriteLine("------------------------------------------------------");
		
		if (args.Length>0){
			foreach(var par in args){
				if (par[0]=='-'){
					if (par=="-ordered"){dither=0;}
					if (par=="-diffuse"){dither=1;}
					if (par=="-nodither"){dither=2;}
				}else{
					// Not an option so probably a bmp file
					fn=par;	
				}
			}
		}else{
			Console.WriteLine("Usage: bmp2cvi [option] myfile.bmp");
			Console.WriteLine("Options: -nodither , -ordered (default), -diffuse");
			return;
		}

		if (File.Exists(fn))
		{
			try
			{
				byte[] bmp_header=new byte[14];
				byte[] bmp_info=new byte[40];
			
				Mybmpinfo bminfo;
				
				Stream infile = File.OpenRead(fn);
				infile.Read(bmp_header,0,bmp_header.Length);
				infile.Read(bmp_info,0,bmp_info.Length);
				
				bminfo.width=swp(bmp_info[4],bmp_info[5],bmp_info[6],bmp_info[7]);
				bminfo.height=swp(bmp_info[8],bmp_info[9],bmp_info[10],bmp_info[11]);
				bminfo.depth=bmp_info[14];
			
				Console.WriteLine("{0} : {1} x {2} x {3}",fn,bminfo.width,bminfo.height,bminfo.depth);
			
				if (bmp_header[0]!='B' && bmp_header[1]!='M'){
					Console.WriteLine("This file does not look like a bitmap image...");
					return;
				}
			
				if ((bminfo.width % 4)!=0){
					Console.WriteLine("The width should be divisible by 4)");
					return;
				}

				if ((bminfo.depth)!=24){
					Console.WriteLine("Images should be 24bit colour.");
					return;
				}
				
				// Got this far so probably a valid bitmap!

				// Now allocate space...
				byte[] mem=new byte[bminfo.width*bminfo.height*3]; 	// File buffer 24bit 320x200x3	
				byte[] img=new byte[bminfo.width*bminfo.height]; 	// Image buffer (8bit gray)
				
				infile.Read(mem,0,mem.Length);				
				infile.Close();
				
				Utils utils=new Utils();

				// Console.WriteLine("Converting 24bit -> 8Bit");
				utils.grey(mem,img,bminfo.width,bminfo.height);
				
				switch(dither){
					case 0:
						Console.WriteLine("Method: Ordered dithering");
						utils.ordered(img,bminfo.width,bminfo.height);
					break;
					
					case 1:
						Console.WriteLine("Method: Diffusing");
						utils.diffuse(img,bminfo.width,bminfo.height);
					break;
					
					case 2:
						Console.WriteLine("Method: No Dither");
						utils.twobit(img,bminfo.width,bminfo.height);
					break;
					
					default:
						Console.WriteLine("Method: Ordered dithering");
						utils.ordered(img,bminfo.width,bminfo.height);
					break;
				}
				
				// Console.WriteLine("Converting to 4 pixels per byte");
				int cy_frame_size = bminfo.width * bminfo.height/4;
				byte[] cy=new byte[cy_frame_size];
				
				utils.frameconv(img,cy,bminfo.width,bminfo.height);

				byte pbyte=popular(cy);
				
				Console.WriteLine("Compression:");

				Cv1 encode=new Cv1();
				
				// Console.WriteLine("Creating fake previous frame");
				
				encode.prev=new byte[cy_frame_size];
				for(int i=0;i<cy_frame_size;i++){
					encode.prev[i]=pbyte;
				}
				
				// See which compression method works the best!
				int maxsize=(bminfo.width*bminfo.height)/4;
									
				int esize=encode.encode_rle_size(cy);
				int usecomp=1;	
				int tmpsize=0;
			
				tmpsize=encode.encode_delta1_size(cy);
				if (tmpsize<esize){
					esize=tmpsize;
					usecomp=2;
				}
				
				tmpsize=encode.encode_delta2_size(cy);
				if (tmpsize<esize){
					esize=tmpsize;
					usecomp=3;
				}
			
				// In case compression just does not work (or not worth it)...
				if (esize>=maxsize){
					esize=maxsize;
					usecomp=0;
				}
				
				// Force 
				// esize=encode_delta2_size(cy,maxsize);
				// usecomp=3;
				
				// esize=encode_rle_size(cy,maxsize);
				// usecomp=1;	
				//
				byte[] cvi_data=new byte[esize];
				
				switch(usecomp)
				{
				case 0:
					// Uncompressed frame
					Console.WriteLine("+ Uncompressed Size: {0} bytes", esize);
					encode.encode_none(cy,cvi_data);
				break;
				
				case 1:
					// RLE compressed frame...
					Console.WriteLine("+ RLE Size: {0} bytes", esize);
					encode.encode_rle(cy,cvi_data);
				break;
				
				case 2:
					// Delta 1 compressed frame...
					Console.WriteLine("+ Delta 1 Size: {0} bytes", esize);
					encode.encode_delta1(cy,cvi_data);
				break;
				
				case 3:
					// Delta 2 compressed frame...
					Console.WriteLine("+ Delta 2 Size: {0} bytes", esize);
					encode.encode_delta2(cy,cvi_data);
				break;
				}
									
				byte[] cvi_header=new byte[20];
				cvi_header[0] =(byte) 'c';
				cvi_header[1] =(byte) 'v';
				cvi_header[2] =(byte) 'i';
				cvi_header[3] =(byte) 'm';

				short swidth = (short)bminfo.width;
				short sheight = (short)bminfo.height;

				// Width
				cvi_header[5] = (byte)(swidth & 0x00ff);
				cvi_header[4] = (byte)((swidth & 0xff00) >> 8);

				// Height
				cvi_header[7] = (byte)(sheight & 0x00ff);
				cvi_header[6] = (byte)((sheight & 0xff00) >> 8);

				cvi_header[8] =(byte)'c';
				cvi_header[9] =(byte)'v';
				cvi_header[10] =(byte)'d';
				cvi_header[11] =(byte)'1';

				// Compression type
				cvi_header[13] = (byte)(usecomp & 0x00ff);
				cvi_header[12] = (byte)((usecomp & 0xff00) >> 8);

				// Most popular byte
				cvi_header[14] = (byte)(pbyte & 0xff);

				// Unused
				cvi_header[15] = 0;

				// Table size
				// const short tablesize= 0;
				cvi_header[17] =0; // (byte)(tablesize & 0x00ff);
				cvi_header[16] =0; // (byte)((tablesize & 0xff00) >> 8);

				//Data size
				cvi_header[19] = (byte)(esize & 0x00ff);
				cvi_header[18] = (byte)((esize & 0xff00) >> 8);
				//==

				string outfn=fn+".cvi";	
				Console.WriteLine("Saving as {0}...",outfn);
				
				using (Stream s = File.Create (outfn)){
					s.Write (cvi_header, 0, cvi_header.Length);
					s.Write (cvi_data, 0, cvi_data.Length);
				}
				
				Console.WriteLine("Finished.");	
			}
			catch (Exception e)
			{
				Console.WriteLine(e);
			}  
		}else{
			Console.WriteLine("Could not find the file [{0}]...",fn);
		}

	return;
	}

}